﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using System.Xml.Linq;
using CloudShot.Core;
using CloudShot.Core.Interfaces;
using CloudShot.Core.Types;
using CloudShot.Utils;

namespace CloudShot
{
	[Serializable]
	public sealed class AppSettings
	{
		#region Static Properties

		private static string _cloudShotAppData;
		private static string CloudShotAppData
		{
			get
			{
				if (string.IsNullOrEmpty(_cloudShotAppData))
					_cloudShotAppData = Application.UserAppDataPath.Replace(Application.ProductVersion, "");

				return _cloudShotAppData;
			}
		}

		private static string _configFile;
		private static string ConfigFile
		{
			get
			{
				if (string.IsNullOrEmpty(_configFile))
					_configFile = Path.Combine(CloudShotAppData, Application.ProductName + @".config");

				return _configFile;
			}
		}

		private static string _errorFile;
		public static string ErrorFile
		{
			get
			{
				if (string.IsNullOrEmpty(_errorFile))
					_errorFile = Path.Combine(CloudShotAppData, @"Error.log");

				return _errorFile;
			}
		}

		private static string _thumbnailsPath;
		public static string ThumbnailsPath
		{
			get
			{
				if (string.IsNullOrEmpty(_thumbnailsPath))
					_thumbnailsPath = Path.Combine(CloudShotAppData, @"Thumbnails");

				return _thumbnailsPath;
			}
		}

		private static string _pluginsPath;
		public static string PluginsPath
		{
			get
			{
				if (string.IsNullOrEmpty(_pluginsPath))
					_pluginsPath = Path.Combine(Application.StartupPath, "Plugins");

				return _pluginsPath;
			}
		}

		#endregion

		#region Custom Application Settings

		public bool StartWithComputer { get; set; }
		public bool OpenInBrowser { get; set; }
		public bool CopyUrl { get; set; }
		public bool CopyImage { get; set; }
		public bool CopyRecentShotPath { get; set; }
		public bool ShowNotifications { get; set; }
		public bool EnableSounds { get; set; }
		public string ShotHotkey { get; set; }
		public bool ShowActiveStorageName { get; set; }
		public Type ActiveStorage { get; set; }

		private ShotInfo[] _recentShotsList = new ShotInfo[0];
		public ShotInfo[] RecentShotsList
		{
			get { return _recentShotsList; }
			set { _recentShotsList = value; }
		}

		private readonly IList<IImageStorage> _storages = null;

		public AppSettings(IList<IImageStorage> storages)
		{
			_storages = storages;
		}

		// Serializes the class to the config file
		// if any of the settings have changed.
		public bool Save()
		{
			UpdateStartupShortcut();

			XDocument doc = new XDocument(
				new XDeclaration("1.0", "utf-8", "yes"),
				//new XComment("CloudShot settings file"),
				new XElement("ApplicationSettings",
					new XElement("StartWithComputer", StartWithComputer),
					new XElement("OpenInBrowser", OpenInBrowser),
					new XElement("CopyUrl", CopyUrl),
					new XElement("CopyImage", CopyImage),
					new XElement("CopyRecentShotPath", CopyRecentShotPath),
					new XElement("ShowNotifications", ShowNotifications),
					new XElement("EnableSounds", EnableSounds),
					new XElement("ShotHotkey", ShotHotkey),
					new XElement("SelectedStorage", ActiveStorage.AssemblyQualifiedName),
					new XElement("ShowActiveStorageName", ShowActiveStorageName),
					new XElement("RecentShots", GetRecentShotList()),
					new XElement("Storages", GetStoragesSettings(_storages))
				)
			);

			doc.Save(ConfigFile);
			return true;
		}

		// Deserializes the class from the config file.
		public bool Load()
		{
			FileInfo fi = new FileInfo(ConfigFile);
			// If the config file exists, open it.
			if (!fi.Exists)
			{
				InitDefaultValues();
				return false;
			}

			XDocument settings = XDocument.Load(fi.FullName);
			XElement data = settings.Descendants("ApplicationSettings").First();

			ShotHotkey = ReadString(data, "ShotHotkey");
			OpenInBrowser = ReadBool(data, "OpenInBrowser");
			CopyUrl = ReadBool(data, "CopyUrl");
			CopyImage = ReadBool(data, "CopyImage");
			CopyRecentShotPath = ReadBool(data, "CopyRecentShotPath");
			StartWithComputer = ReadBool(data, "StartWithComputer");
			ShowNotifications = ReadBool(data, "ShowNotifications");
			EnableSounds = ReadBool(data, "EnableSounds");
			ShowActiveStorageName = ReadBool(data, "ShowActiveStorageName");
			ActiveStorage = ReadType(data, "SelectedStorage");
			RecentShotsList = ReadRecentShots(data);
			LoadStorageSettings(data, _storages);

			UpdateStartupShortcut();
			return true;
		}

		private void InitDefaultValues()
		{
			try
			{
				StartWithComputer = ReadSettingBool("StartWithComputer", true);
				OpenInBrowser = ReadSettingBool("OpenInBrowser", true);
				CopyUrl = ReadSettingBool("CopyUrl", true);
				CopyImage = ReadSettingBool("CopyImage", false);
				CopyRecentShotPath = ReadSettingBool("CopyRecentShotPath", false);
				ShowNotifications = ReadSettingBool("ShowNotifications", true);
				EnableSounds = ReadSettingBool("EnableSounds", true);
				ShowActiveStorageName = ReadSettingBool("ShowActiveStorageName", true);

				string selectedStorageType = ConfigurationManager.AppSettings["SelectedStorage"];

				if (!string.IsNullOrEmpty(selectedStorageType))
					ActiveStorage = ReadType(selectedStorageType);

				string hotkey = ConfigurationManager.AppSettings["ShotHotkey"];
				ShotHotkey = string.IsNullOrEmpty(hotkey) ? Keys.PrintScreen.ToString() : hotkey;

				foreach (IImageStorage storage in _storages)
				{
					if (storage.SupportedStorageType == StorageType.CustomLocal || storage.SupportedStorageType == StorageType.CustomRemote)
					{
						var folderToSave = ConfigurationManager.AppSettings["FolderToSave"];

						//TODO: at the current moment initial value for FolderToSave set in IImageStorage, but maybe it's better to set this folder here if it is empty after config applying.
						if (!string.IsNullOrEmpty(folderToSave))
							storage.FolderToSave = folderToSave;
					}
				}
			}
			catch { }
		}

		private void LoadStorageSettings(XElement data, IList<IImageStorage> imageStorages)
		{
			XElement element = data.Element("Storages");
			if (element != null)
			{
				var q = from x in element.Elements()
								where !x.IsEmpty
								select x;

				foreach (XElement xStorage in q)
				{
					foreach (IImageStorage imageStorage in imageStorages)
					{
						var el = GetStorageElement(imageStorage);
						if (xStorage.Name == el.Name)
						{
							var xStorageAttr = xStorage.Attribute("PluginDir");
							var elAttr = el.Attribute("PluginDir");

							if ((xStorageAttr == null && elAttr == null))
								imageStorage.LoadSettings(xStorage.Elements().ToArray());
							else if (xStorageAttr != null && elAttr != null && xStorageAttr.Value == elAttr.Value)
								imageStorage.LoadSettings(xStorage.Elements().ToArray());
						}
					}
				}
			}
		}

		private ShotInfo[] ReadRecentShots(XElement element)
		{
			XElement recentShots = element.Element("RecentShots");

			if (recentShots == null)
				return new ShotInfo[0];

			var q = from c in recentShots.Elements("Shot")
							let shot = ReadShot(c)
							where shot != null
							select shot;

			ShotInfo[] result = q.ToArray();

			return result;
		}

		private static ShotInfo ReadShot(XElement element)
		{
			ShotInfo result = null;

			if (element != null && !element.IsEmpty && element.HasElements)
			{
				result = new ShotInfo
				{
					Name = ReadString(element, "Name"),
					Path = ReadString(element, "Path")
				};
			}

			return result;
		}

		private XElement[] GetStoragesSettings(IList<IImageStorage> storages)
		{
			XElement[] result = new XElement[storages.Count];

			for (int i = 0; i < storages.Count; i++)
			{
				IImageStorage storage = storages[i];
				XElement element = GetStorageElement(storage);

				XElement[] setings = storage.SaveSettings();
				element.Add(setings);
				result[i] = element;

				//MessageBox.Show(storage.GetType().Module.FullyQualifiedName);
				//MessageBox.Show(storage.GetType().Module.FullyQualifiedName);
				//result[i] = storage.SaveSettings();
			}

			return result;
		}

		private XElement GetStorageElement(IImageStorage storage)
		{
			Type storageType = storage.GetType();
			string path = storageType.Module.FullyQualifiedName;

			if (string.Equals(path, Application.ExecutablePath, StringComparison.InvariantCultureIgnoreCase))
			{
				path = null;
			}
			else
			{
				path = Path.GetDirectoryName(path) ?? string.Empty;
				path = path.Replace(PluginsPath, "").Trim('\\', '/');
			}

			var result = new XElement(storageType.FullName);

			if (!string.IsNullOrEmpty(path))
				result.SetAttributeValue("PluginDir", path);

			return result;
		}

		private XElement[] GetRecentShotList()
		{
			XElement[] result = new XElement[RecentShotsList.Length];

			for (int i = 0; i < RecentShotsList.Length; i++)
			{
				ShotInfo shot = RecentShotsList[i];

				result[i] = new XElement("Shot",
					//new XAttribute("Name", shot.Name),
					new XElement("Name", shot.Name),
					new XElement("Path", shot.Path)
					);
			}

			return result;
		}

		#endregion

		#region Helper Methods

		private void UpdateStartupShortcut()
		{
			RegistryUtils.SetStartup(Application.ProductName, StartWithComputer);
		}

		private bool ReadSettingBool(string name, bool defaultValue)
		{
			bool result;

			if (bool.TryParse(ConfigurationManager.AppSettings[name], out result))
				return result;

			return defaultValue;
		}

		private static Type ReadType(XElement element, string name)
		{
			return ReadType(ReadString(element, name));
		}

		private static Type ReadType(string value)
		{
			Type result = null;

			try
			{
				result = Type.GetType(value, false);
			}
			catch (Exception err)
			{
				ErrorsHelper.TraceError(err);
			}

			return result;
		}

		private static bool ReadBool(XElement element, string name)
		{
			bool result = false;
			bool.TryParse(ReadString(element, name), out result);
			return result;
		}

		private static string ReadString(XElement element, string name)
		{
			if ((element == null) || (element.Element(name) == null))
				return string.Empty;

			return element.Element(name).Value;
		}

		#endregion
	}
}