﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Media;
using System.Windows.Forms;
using CloudShot.Core;
using CloudShot.Core.Interfaces;
using CloudShot.Core.Types;
using CloudShot.ImageStorages;
using CloudShot.Properties;
using CloudShot.Utils;
using DropboxImageStorage;

namespace CloudShot
{
	public partial class MainForm : Form
	{
		private const int UPDATE_INTERVAL_HOURS = 8;
		private const string DONATE_URL = @"http://cloudshot.codeplex.com/wikipage?title=Donate";

		private readonly GlobalHotListener _hotkeyListener;
		private readonly AppSettings _appSettings;
		private readonly RecentShotsQueue _recentShots = new RecentShotsQueue();

		private readonly List<IImageStorage> _storages = new List<IImageStorage>();

		private IImageStorage _activeStorage;
		private IImageStorage ActiveStorage
		{
			get { return _activeStorage; }
			set
			{
				_activeStorage = value;
				OnActiveStorageChanged();
			}
		}

		private Hotkey _hotkeyShot;
		private bool _isShottingStarted = false;

		private string ProductAndVersion { get { return Application.ProductName + " " + Application.ProductVersion; } }

		public MainForm()
		{
			InitializeComponent();

			this.Icon = Resources.applicationIcon;
			treeIcon.Icon = Resources.applicationIcon;
			treeIcon.Text = ProductAndVersion;

			_hotkeyListener = new GlobalHotListener(this);

			_appSettings = new AppSettings(_storages);
			InitStorages(_appSettings);
		}

		private void InitStorages(AppSettings settings)
		{
			_storages.Add(new FileSystemStorage());
			_storages.Add(new DropboxStorage());


			var pm = new PluginManager(AppSettings.PluginsPath);

			foreach (var plugin in pm.GetPlugins())
			{
				_storages.Add(plugin);
			}


			//Load storages settings
			_appSettings.Load();

			//Set active storage
			ActiveStorage = _storages.FirstOrDefault(s => s.GetType() == settings.ActiveStorage);

			if (ActiveStorage == null)
			{
				ActiveStorage = _storages.First();
				_appSettings.ActiveStorage = ActiveStorage.GetType();
			}

			InitStoragesMenu();
		}

		private void InitStoragesMenu()
		{
			if (_storages.Count < 2)
				return;

			int index = treeIcon.ContextMenuStrip.Items.IndexOf(mnuSettings);

			if (index == -1)
				return;

			//add storages to menu
			foreach (IImageStorage storage in _storages)
			{
				var item = new ToolStripMenuItem(storage.Name, storage.Logo);
				item.Tag = storage;
				item.Checked = storage == ActiveStorage;
				item.Click += OnStorageItemClick;

				treeIcon.ContextMenuStrip.Items.Insert(index++, item);
			}
			treeIcon.ContextMenuStrip.Items.Insert(index, new ToolStripSeparator());
		}

		//TODO переделать логику работы, хоткей должен проверяться в Settings и при первом запуске
		private void SetupHotkey()
		{
			if (this.IsDisposed)
				return;

			//удаляем старые хоткеи, если они уже были заданы
			if (_hotkeyShot != null)
			{
				_hotkeyShot.HotkeyPressed -= OnShot;
				_hotkeyListener.Remove(_hotkeyShot);
			}

			if (!Hotkey.TryParse(_appSettings.ShotHotkey, out _hotkeyShot))
			{
				_appSettings.ShotHotkey = string.Empty;
				_appSettings.Save();
				return;
			}

			if (!_hotkeyListener.Add(_hotkeyShot))
			{
				string hotkeyName = _hotkeyShot.ToString();

				_hotkeyShot = null;
				_appSettings.ShotHotkey = string.Empty;
				_appSettings.Save();

				MessageBox.Show(string.Format("Hotkey \"{0}\" used by another application.\nPlease select another hotkey in Settings.", hotkeyName),
					Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
			}

			if (_hotkeyShot != null)
			{
				_hotkeyShot.HotkeyPressed += OnShot;
				mnuShot.ShortcutKeyDisplayString = _hotkeyShot.ToString();
			}
		}

		#region Recent Shots

		private void FillRecentShotsMenu(IEnumerable<ShotInfo> shots)
		{
			mnuRecentShots.DropDownItems.Clear();

			foreach (ShotInfo recentShot in shots)
			{
				ToolStripMenuItem menuItem = new ToolStripMenuItem();
				menuItem.ImageScaling = ToolStripItemImageScaling.None;
				menuItem.Text = recentShot.Name;
				menuItem.Tag = recentShot;
				menuItem.Image = recentShot.Thumbnail;
				menuItem.Click += OnRecentShotClick;

				mnuRecentShots.DropDownItems.Insert(0, menuItem);
			}

			mnuRecentShots.Enabled = mnuRecentShots.DropDownItems.Count > 0;
		}

		private void OnRecentShotClick(object sender, EventArgs e)
		{
			ToolStripMenuItem menu = sender as ToolStripMenuItem;

			if (menu == null)
				return;

			ShotInfo recentShot = menu.Tag as ShotInfo;

			if (recentShot == null)
				return;

			OpenInViewer(recentShot);

			if (_appSettings.CopyRecentShotPath)
			{
				if (SetClipboardData(recentShot, null))
					ShowNotification(recentShot, "Link copied to clipboard");
			}
		}

		private void AddRecentShot(ShotInfo shotInfo)
		{
			_recentShots.Add(shotInfo);

			FillRecentShotsMenu(_recentShots);

			_appSettings.RecentShotsList = _recentShots.ToArray();
			_appSettings.Save();
		}

		#endregion

		private void Shot()
		{
			if (_isShottingStarted)
				return;

			if (ActiveStorage == null)
				return;

			try
			{
				_isShottingStarted = true;

				Image image = MakeScreenshot();

				if (image == null)
					return;

				//authorize
				if (!ActiveStorage.IsAuthorized)
				{
					if (MessageBox.Show(Resources.AuthorizationQuestion, Application.ProductName, MessageBoxButtons.YesNo, MessageBoxIcon.Information) != DialogResult.Yes)
						return;

					ActiveStorage.SignInAsync(success =>
					{
						if (success)
						{
							_appSettings.Save();
							SaveScreenshot(image);
						}
					});

					return;
				}

				SaveScreenshot(image);
			}
			catch (Exception err)
			{
				ErrorsHelper.TraceError(err);
				ErrorsHelper.ShowError(err);
			}
			finally
			{
				_isShottingStarted = false;
			}
		}

		private Image MakeScreenshot()
		{
			using (var shotForm = new ShotForm(_appSettings.ShowActiveStorageName ? ActiveStorage.Name : null))
			{
				if (shotForm.ShowDialog() != DialogResult.OK || shotForm.Image == null)
					return null;

				if (_appSettings.EnableSounds)
					PlayShotSound();

				return shotForm.Image;
			}
		}

		private void SaveScreenshot(Image image)
		{
			Clipboard.Clear();//clear old screenshot url from clipboard

			Action<ShotInfo> saveAction = null;

			if (_appSettings.OpenInBrowser)
				saveAction = s => this.Invoke((MethodInvoker)(() => OpenInViewer(s)));

			ShotInfo shotInfo = ActiveStorage.SaveImageAsync(image, AppHelpers.GetFileName(ActiveStorage.CurrentImageFormat), saveAction);

			AddRecentShot(shotInfo);
			SetClipboardData(shotInfo, image);
			ShowNotification(shotInfo, "Shot saved to storage");

			image.Dispose();
			image = null;
		}

		private void ShowNotification(ShotInfo shotInfo, string text)
		{
			if (!_appSettings.ShowNotifications)
				return;

			treeIcon.Tag = shotInfo;
			treeIcon.ShowBalloonTip(1, text, shotInfo.Path, ToolTipIcon.Info);
		}

		private void PlayShotSound()
		{
			using (var player = new SoundPlayer(Resources.CAMERA))
			{
				player.Play();
			}
		}

		private void SetTrayTooltip()
		{
			string hotkey = "Not set";

			if (_hotkeyShot != null)
				hotkey = _hotkeyShot.ToString();

			string treeIconText = string.Format("{0}{1}Hotkey: {2}{1}Storage: {3}", ProductAndVersion, Environment.NewLine, hotkey, ActiveStorage.Name);

			const int maxTaryTextLenght = 63;

			if (treeIconText.Length > maxTaryTextLenght)
				treeIconText = treeIconText.Substring(0, maxTaryTextLenght);

			treeIcon.Text = treeIconText;
		}

		private bool SetClipboardData(ShotInfo shot, Image screenshot)
		{
			//Debug.Print("Copy shot to clipboard: " + url);

			if (_appSettings.CopyUrl || _appSettings.CopyImage)
			{
				//Clipboard.Clear();
				var dataObject = new DataObject();
				if (_appSettings.CopyUrl && !string.IsNullOrEmpty(shot.Path)) dataObject.SetText(shot.Path);
				if (_appSettings.CopyImage && screenshot != null) dataObject.SetImage(screenshot);
				Clipboard.SetDataObject(dataObject, true, 4, 500);
				return true;
			}

			return false;
		}

		private void OpenInViewer(ShotInfo shotInfo)
		{
			if (ActiveStorage == null)
				return;

			try
			{
				Process.Start(shotInfo.Path);
			}
			catch (Exception err)
			{
				//RemoveRecentShot(shotInfo);//TODO продумать как это сделать
				ErrorsHelper.ShowError(err);
			}
		}

		private void ShowSettings()
		{
			mnuSettings.Enabled = false;

			using (var dialog = new SettingsForm(_appSettings, _storages, ActiveStorage))
			{
				if (dialog.ShowDialog(this) == DialogResult.OK)
					ActiveStorage = dialog.ActiveStorage;
			}

			SetupHotkey(); //в любом случае задаем хоткеи, т.к. они были стерты через ClearHotkeys
			SetTrayTooltip();
			mnuSettings.Enabled = true;
		}

		private void TryUpdateApplication()
		{
			ApplicationUpdater.InstallUpdateIfAvailable(this);
		}

		#region Event Handlers

		private void OnLoad(object sender, EventArgs e)
		{
			this.Visible = false;

			SetupHotkey();
			SetTrayTooltip();
			_recentShots.Load(_appSettings.RecentShotsList);
			FillRecentShotsMenu(_recentShots);

			treeIcon.Visible = true;

			var interval = new TimeSpan(UPDATE_INTERVAL_HOURS, 0, 0);//check updates every UPDATE_INTERVAL_HOURS hours
			tmrUpdater.Interval = (int)interval.TotalMilliseconds;
			tmrUpdater.Enabled = true;
			tmrUpdater.Start();
			TryUpdateApplication();
		}

		private void OnShot(object sender, EventArgs e)
		{
			Shot();
		}

		private void OnMenuExitClick(object sender, EventArgs e)
		{
			Application.Exit();
		}

		private void OnTreeIconMouseDoubleClick(object sender, MouseEventArgs e)
		{
			if (mnuShot.Enabled)
			{
				Shot();
			}
			else
			{
				var form = this.OwnedForms.FirstOrDefault();
				if (form != null)
					form.Activate();
			}
		}

		private void OnMenuSettingsClick(object sender, EventArgs e)
		{
			ShowSettings();
		}

		private void OnMenuDonateClick(object sender, EventArgs e)
		{
			if (string.IsNullOrEmpty(DONATE_URL))
				return;

			Process.Start(DONATE_URL);
			this.Close();
		}

		private void OnMenuAboutClick(object sender, EventArgs e)
		{
			mnuAbout.Enabled = false;

			using (var dialog = new AboutBox())
			{
				dialog.ShowDialog();
			}

			mnuAbout.Enabled = true;
		}

		private void OnTimerTick(object sender, EventArgs e)
		{
			TryUpdateApplication();
		}

		private void OnStorageItemClick(object sender, EventArgs e)
		{
			ToolStripMenuItem item = sender as ToolStripMenuItem;

			if (item == null)
				return;

			IImageStorage storage = item.Tag as IImageStorage;

			if (storage != null)
			{
				ActiveStorage = storage;
				_appSettings.ActiveStorage = storage.GetType();
				_appSettings.Save();
				SetTrayTooltip();
			}
		}

		private void OnActiveStorageChanged()
		{
			foreach (ToolStripItem item in treeIcon.ContextMenuStrip.Items)
			{
				if (item is ToolStripMenuItem)
					((ToolStripMenuItem)item).Checked = item.Tag == ActiveStorage;
			}
		}

		#endregion

		private void OnTreeIconBalloonTipClicked(object sender, EventArgs e)
		{
			if (treeIcon.Tag is ShotInfo)
			{
				OpenInViewer((ShotInfo)treeIcon.Tag);
				treeIcon.Tag = null;
			}
		}
	}
}