﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using CloudShot.Core;
using CloudShot.Core.Interfaces;
using CloudShot.Core.Types;
using CloudShot.Core.Utils;
using CloudShot.Properties;

namespace CloudShot
{
	public partial class SettingsForm : Form
	{
		private readonly AppSettings _appSettings;
		private readonly List<IImageStorage> _storages;
		public IImageStorage ActiveStorage { get; private set; }

		public SettingsForm(AppSettings appSettings, IEnumerable<IImageStorage> storages, IImageStorage activeStorage)
		{
			_appSettings = appSettings;
			_storages = new List<IImageStorage>(storages);

			InitializeComponent();
			this.Icon = Resources.applicationIcon;

			LoadSettings(activeStorage);
		}

		private void OnOkClick(object sender, EventArgs e)
		{
			if (ActiveStorage == null)
			{
				MessageBox.Show("Not set storage.\nPlease select storage.", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				tabContainer.SelectTab(tabStorage);
				cboStorages.Focus();
				return;
			}

			if (cboImageFormat.SelectedItem == null)
			{
				MessageBox.Show("Not set image format.\nPlease select image format.", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				tabContainer.SelectTab(tabStorage);
				cboImageFormat.Focus();
				return;
			}

			if (ActiveStorage.SupportedStorageType == StorageType.CustomLocal)
			{
				//TODO по иджее надо делать такую же проверку при шоте
				try
				{
					Directory.CreateDirectory(txtPath.Text);
				}
				catch (Exception err)
				{
					string message = string.Format("Invalid path to save.\n{0}\nPlease enter correct path.", err.Message);
					MessageBox.Show(message, Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
					tabContainer.SelectTab(tabStorage);
					txtPath.Focus();
					return;
				}
			}

			ActiveStorage.FolderToSave = txtPath.Text;
			Debug.Assert(cboImageFormat.SelectedItem != null);
			ActiveStorage.CurrentImageFormat = ImageTypeUtils.GetImageType((string)cboImageFormat.SelectedItem);

			SaveSettings();

			DialogResult = DialogResult.OK;
		}

		private void LoadSettings(IImageStorage activeStorage)
		{
			chkOpenInBrowser.Checked = _appSettings.OpenInBrowser;
			chkCopyUrl.Checked = _appSettings.CopyUrl;
			chkCopyImage.Checked = _appSettings.CopyImage;
			chkCopyRecentShotPath.Checked = _appSettings.CopyRecentShotPath;
			hkbRegion.Text = _appSettings.ShotHotkey;
			chkNotifications.Checked = _appSettings.ShowNotifications;
			chkStartWithWindows.Checked = _appSettings.StartWithComputer;
			chkEnableSounds.Checked = _appSettings.EnableSounds;
			chkShowActiveStorageName.Checked = _appSettings.ShowActiveStorageName;


			cboStorages.DisplayMember = "Name";
			cboStorages.DataSource = _storages;
			cboStorages.SelectedItem = activeStorage;
		}

		private void SaveSettings()
		{
			_appSettings.OpenInBrowser = chkOpenInBrowser.Checked;
			_appSettings.CopyUrl = chkCopyUrl.Checked;
			_appSettings.CopyImage = chkCopyImage.Checked;
			_appSettings.CopyRecentShotPath = chkCopyRecentShotPath.Checked;
			_appSettings.ShotHotkey = hkbRegion.Text;
			_appSettings.ShowNotifications = chkNotifications.Checked;
			_appSettings.EnableSounds = chkEnableSounds.Checked;
			_appSettings.ShowActiveStorageName = chkShowActiveStorageName.Checked;

			_appSettings.StartWithComputer = chkStartWithWindows.Checked;

			if (ActiveStorage != null)
				_appSettings.ActiveStorage = ActiveStorage.GetType();

			_appSettings.Save();
		}

		private void OnBrowseClick(object sender, EventArgs e)
		{
			var dialog = new FolderBrowserDialog();
			dialog.Description = "Select folder to save screenshots";

			if (!string.IsNullOrEmpty(txtPath.Text))
				dialog.SelectedPath = txtPath.Text;

			if (dialog.ShowDialog() == DialogResult.OK)
			{
				txtPath.Text = dialog.SelectedPath;
				txtPath.SelectionStart = txtPath.Text.Length;
				toolTip.SetToolTip(txtPath, txtPath.Text);
			}

			//BrowseToFolder("dddd", DropboxUtils.DropboxPublicFolder);
		}

		///// <summary>
		///// Invokes the system BrowseToFolder, but allowing a user-specified root path
		///// </summary>
		///// <param name="title">The title for the dialog</param>
		///// <param name="rootPath">The root path where browsing starts</param>
		///// <returns>The user-selected path if the user pressed OK, or null if the user cancelled</returns>
		//public string BrowseToFolder(string title, string rootPath)
		//{
		//  Type shellType = Type.GetTypeFromProgID("Shell.Application");
		//  object shell = Activator.CreateInstance(shellType);
		//  object folder = shellType.InvokeMember("BrowseForFolder", BindingFlags.InvokeMethod, null, shell, new object[] { 0, title, 0, rootPath });

		//  if (folder == null)
		//    return null; // User clicked cancel

		//  object folderSelf = folder.GetType().InvokeMember("Self", BindingFlags.GetProperty, null, folder, null);

		//  return folderSelf.GetType().InvokeMember("Path", BindingFlags.GetProperty, null, folderSelf, null) as string;
		//}

		private void OnActiveStorageChanged(object sender, EventArgs e)
		{
			ActiveStorage = cboStorages.SelectedItem as IImageStorage;

			if (ActiveStorage != null)
				ShowStorageSettings(ActiveStorage);
		}

		private void ShowStorageSettings(IImageStorage storage)
		{
			//Image Formats
			cboImageFormat.Items.Clear();

			foreach (string extension in storage.SupportedFormats.GetImageExtensions())
			{
				cboImageFormat.Items.Add(extension);
			}

			if (storage.SupportedFormats.Contains(storage.CurrentImageFormat))
				cboImageFormat.SelectedItem = storage.CurrentImageFormat.GetImageExtension();
			else
				cboImageFormat.SelectedItem = storage.SupportedFormats.FirstOrDefault().GetImageExtension();

			//Path to Save
			txtPath.Text = storage.FolderToSave;
			txtPath.SelectionStart = txtPath.Text.Length;
			toolTip.SetToolTip(txtPath, txtPath.Text);

			if (storage.AuthorizationMode == AuthorizationMode.OAuth)
			{
				oAuthControl.Visible = true;
				oAuthControl.SetState(storage.IsAuthorized ? AuthState.Authorized : AuthState.NotAuthorized);

				cboImageFormat.Visible = oAuthControl.AuthState == AuthState.Authorized;
				lblImageFormat.Visible = oAuthControl.AuthState == AuthState.Authorized;
				pnlSaveToFolder.Visible = (storage.SupportedStorageType == StorageType.CustomLocal || storage.SupportedStorageType == StorageType.CustomRemote) &&
					oAuthControl.AuthState == AuthState.Authorized;
				btnBrowse.Visible = storage.SupportedStorageType == StorageType.CustomLocal && oAuthControl.AuthState == AuthState.Authorized;
			}
			else
			{
				oAuthControl.Visible = false;

				cboImageFormat.Visible = true;
				lblImageFormat.Visible = true;
				pnlSaveToFolder.Visible = storage.SupportedStorageType == StorageType.CustomLocal || storage.SupportedStorageType == StorageType.CustomRemote;
				btnBrowse.Visible = storage.SupportedStorageType == StorageType.CustomLocal;

			}

			toolTip.SetToolTip(cboStorages, storage.Description);
			toolTip.SetToolTip(picInfo, storage.Description);
		}

		private void Unlink()
		{
			ActiveStorage.SignOut();
			SaveSettings();
			oAuthControl.SetState(AuthState.NotAuthorized);
			ShowStorageSettings(ActiveStorage);
		}

		private void OnSignInClick(object sender, EventArgs e)
		{
			ActiveStorage.SignInAsync(OnAuthorized);
			oAuthControl.SetState(AuthState.WaitAuthorization);
		}

		private void OnAuthorized(bool success)
		{
			SaveSettings();
			oAuthControl.SetState(success ? AuthState.Authorized : AuthState.NotAuthorized);
			ShowStorageSettings(ActiveStorage);
		}

		private void OnUnlinkClick(object sender, EventArgs e)
		{
			Unlink();
		}

		private void OnCancelWaitingClick(object sender, EventArgs e)
		{
			Unlink();
		}

		private void OnClosing(object sender, FormClosingEventArgs e)
		{
			//cancel wait to authorization
			if (oAuthControl.AuthState == AuthState.WaitAuthorization)
				ActiveStorage.SignOut();
		}
	}
}