﻿using System;
using System.Drawing;
using System.Drawing.Text;
using System.Windows.Forms;
using CloudShot.Properties;
using CloudShot.Utils;

namespace CloudShot.States
{
	class ShotState : IShotState
	{
		private readonly Form _owner;
		private readonly string _storageName;
		private readonly Pen _borderPen = new Pen(Color.FromArgb(100, 100, 100), 1);

		private bool _pressed;

		private bool _selectionMoveEnabled;
		private bool _startXGreater;
		private bool _startYGreater;

		private bool _ctrlPressed;
		private Point _start;
		private Point _end;
		private Rectangle _selectionRect;
		private bool _filterEnabled = false;

		private readonly OpenWindowsManager _windowManager = new OpenWindowsManager();
		private readonly Bitmap _lightBitmap;
		private string _typedText = string.Empty;
		private readonly Font _typedTextFontBig;
		private readonly Font _typedTextFontSmall;
		private readonly Font _storageFont;
		private readonly SolidBrush _fillBrush;


		public ShotState(Form owner, string storageName)
		{
			_owner = owner;
			_storageName = storageName;

			_typedTextFontBig = new Font(_owner.Font.FontFamily, 15f);
			_typedTextFontSmall = new Font(_owner.Font.FontFamily, 11f, FontStyle.Regular);
			_storageFont = new Font(_owner.Font.FontFamily, 150f, FontStyle.Regular);
			_fillBrush = new SolidBrush(Color.FromArgb(190, Color.Black));

			_lightBitmap = new Bitmap(_owner.BackgroundImage);
			Bitmap darkBitmap = new Bitmap(_lightBitmap);

			using (Graphics g = Graphics.FromImage(darkBitmap))
			{
				g.FillRectangle(_fillBrush, 0, 0, darkBitmap.Width, darkBitmap.Height);
			}

			_owner.BackgroundImage = darkBitmap;

			//subscribe
			_owner.KeyDown += OnKeyDown;
			_owner.KeyUp += OnKeyUp;
			_owner.KeyPress += OnKeyPress;
		}

		#region Event Handlers

		private void OnKeyDown(object sender, KeyEventArgs e)
		{
			if (_pressed && e.KeyCode == Keys.Space && !_selectionMoveEnabled)
			{
				_selectionMoveEnabled = true;
				_startXGreater = _end.X < _start.X;
				_startYGreater = _end.Y < _start.Y;
				return;
			}

			if (e.KeyCode == Keys.V && e.Modifiers == Keys.Control)
			{
				string text = Clipboard.GetText();
				if (!string.IsNullOrEmpty(text))
				{
					_typedText += text;
				}
			}
			else
			{
				_ctrlPressed = e.Control;
				_owner.Invalidate();
			}
		}

		private void OnKeyUp(object sender, KeyEventArgs e)
		{
			if (_selectionMoveEnabled)
			{
				_start.X = (_end.X == _selectionRect.Location.X) ? _end.X + _selectionRect.Width : _selectionRect.Location.X;
				_start.Y = (_end.Y == _selectionRect.Location.Y) ? _end.Y + _selectionRect.Height : _selectionRect.Location.Y;
				_selectionMoveEnabled = false;
			}

			_ctrlPressed = false;

			if (e.KeyData == Keys.Menu)
				_filterEnabled = !_filterEnabled;

			_owner.Invalidate();
		}

		private void OnKeyPress(object sender, KeyPressEventArgs e)
		{
			if (_pressed)
				return;

			if (e.KeyChar == '\b')//backspace
			{
				if (!string.IsNullOrEmpty(_typedText))
				{
					_typedText = _typedText.Remove(_typedText.Length - 1);
				}
			}
			else if (char.IsLetterOrDigit(e.KeyChar) || char.IsPunctuation(e.KeyChar) || char.IsWhiteSpace(e.KeyChar))
			{
				_typedText += e.KeyChar;
			}
		}

		#endregion

		#region Drawing

		public void Paint(Graphics graphics)
		{
			Point screenCursorPos = Cursor.Position;
			Screen currentScreen = Screen.FromPoint(screenCursorPos);
			Rectangle screenClientBounds = _owner.RectangleToClient(currentScreen.Bounds);

			_end = _owner.PointToClient(screenCursorPos);

#if DEBUG
			WindowInfo wi = _windowManager.GetWindowRectFromPoint(_end, new IntPtr[] { _owner.Handle });

			if (wi != null)
			{
				graphics.DrawRectangle(Pens.Red, wi.Bounds);
				graphics.DrawRectangle(Pens.DimGray, _owner.RectangleToClient(wi.Bounds));
			}
#endif

			if (_pressed)
			{
				RecalculateSelection();
				Rectangle rect = _selectionRect;

				if (rect.Width <= 0 || rect.Height <= 0)
					return;

				graphics.DrawImage(_lightBitmap, rect.X, rect.Y, rect, GraphicsUnit.Pixel);
				if (!string.IsNullOrEmpty(_typedText))
					DrawTypedText(graphics, rect);
				graphics.DrawRectangle(_borderPen, rect);

				//Text label
				string label = string.Format("{0} x {1}", rect.Width, rect.Height);
				
				//show label on right bottom corner of selection rectangle
				DrawTextLabel(label, new Point(rect.Right + 1, rect.Bottom + 1), graphics, false);

				//show size label on center
				//SizeF labelSize = TextRenderer.MeasureText(label, _owner.Font);
				//if (rect.Width <= labelSize.Width || rect.Height <= labelSize.Height)
				//{
				//  DrawTextLabel(label, new Point(rect.Right, rect.Bottom), graphics, false);
				//}
				//else
				//{
				//  graphics.FillRectangle(_fillBrush, rect.Left + (rect.Width / 2 - labelSize.Width / 2), rect.Top + (rect.Height / 2 - labelSize.Height / 2), labelSize.Width, labelSize.Height);
				//  TextRenderer.DrawText(graphics, label, _owner.Font, rect, Color.WhiteSmoke, TextFormatFlags.HorizontalCenter | TextFormatFlags.VerticalCenter | TextFormatFlags.LeftAndRightPadding);
				//}

				if (!_selectionMoveEnabled)
					DrawTip(graphics, Resources.TipHoldSpace, screenClientBounds);
			}
			else
			{
				string label = string.Format("{0} x {1}", screenCursorPos.X - currentScreen.Bounds.X, screenCursorPos.Y - currentScreen.Bounds.Y);

				DrawTextLabel(label, _end, graphics, true);

				if (_ctrlPressed)
					DrawCross(graphics);

				if (string.IsNullOrEmpty(_typedText) && string.IsNullOrEmpty(_storageName))
					return;

				if (!string.IsNullOrEmpty(_typedText))
					TextRenderer.DrawText(graphics, _typedText, _typedTextFontBig, screenClientBounds, Color.WhiteSmoke, TextFormatFlags.WordBreak | TextFormatFlags.HorizontalCenter);

				if (!string.IsNullOrEmpty(_storageName))
				{
					SizeF size = graphics.MeasureString(_storageName, _storageFont, screenClientBounds.Width);
					graphics.DrawString(_storageName, _storageFont, new SolidBrush(Color.FromArgb(5, Color.WhiteSmoke)), (screenClientBounds.Width - size.Width) / 2, (screenClientBounds.Height - size.Height) / 2);
				}
			}
		}

		private void DrawTextLabel(string label, Point pos, Graphics graphics, bool onRightSide)
		{
			SizeF labelSize = TextRenderer.MeasureText(label, _owner.Font);

			int x;

			if (onRightSide)
				x = (pos.X + labelSize.Width < _owner.Width) ? pos.X : pos.X - (int)labelSize.Width;
			else
				x = Math.Max(pos.X - (int)labelSize.Width, 0);

			int y = (pos.Y + labelSize.Height < _owner.Height) ? pos.Y : pos.Y - (int)labelSize.Height;

			if ((pos.Y > y) && (pos.X + labelSize.Width < _owner.Width))
				x = pos.X;

			Rectangle labelRect = new Rectangle(x, y, (int)labelSize.Width, (int)labelSize.Height);
			graphics.FillRectangle(Brushes.WhiteSmoke, labelRect);
			TextRenderer.DrawText(graphics, label, _owner.Font, labelRect, Color.Black, TextFormatFlags.HorizontalCenter | TextFormatFlags.VerticalCenter | TextFormatFlags.LeftAndRightPadding);
		}

		private void DrawTypedText(Graphics graphics, Rectangle rect)
		{
			graphics.TextRenderingHint = TextRenderingHint.ClearTypeGridFit;

			//calculating
			Size size = TextRenderer.MeasureText(_typedText, _typedTextFontSmall, rect.Size, TextFormatFlags.WordBreak);
			Rectangle textRect = new Rectangle(rect.X, rect.Bottom - size.Height, rect.Width, size.Height);

			//drawing
			graphics.FillRectangle(new SolidBrush(Color.FromArgb(200, Color.Black)), textRect);
			TextRenderer.DrawText(graphics, _typedText, _typedTextFontSmall, textRect, Color.WhiteSmoke, TextFormatFlags.WordBreak | TextFormatFlags.HorizontalCenter);
		}

		private void DrawCross(Graphics graphics)
		{
			using (Pen pen = new Pen(Color.FromArgb(180, Color.Gray)))
			{
				Point pos = _owner.PointToClient(Cursor.Position);
				graphics.DrawLine(pen, pos.X, 0, pos.X, _owner.Height);
				graphics.DrawLine(pen, 0, pos.Y, _owner.Width, pos.Y);
			}
		}

		private void DrawTip(Graphics graphics, string text, Rectangle rect)
		{
			graphics.TextRenderingHint = TextRenderingHint.ClearTypeGridFit;

			//calculating
			Size size = TextRenderer.MeasureText(text, _typedTextFontSmall, new Size(500, rect.Height), TextFormatFlags.WordBreak);

			int x = (rect.Width - size.Width) / 2;
			int y = rect.Height - 100;
			Rectangle textRect = new Rectangle(x, y, size.Width, size.Height);

			if (_selectionRect.IntersectsWith(textRect))
				return;

			TextRenderer.DrawText(graphics, text, _typedTextFontSmall, textRect, Color.Black, Color.LightGoldenrodYellow, TextFormatFlags.HorizontalCenter);
		}

		#endregion

		#region IShotState

		public void Start(Point location)
		{
			_start = location;
			_pressed = true;
		}

		public virtual void Update(Point location)
		{
		}

		public void End()
		{
			_pressed = false;
		}

		public Image Shot()
		{
			Rectangle shotRect = Rectangle.Empty;

			if (_start == _end)
			{
				//get window shot
				Point screenPoint = _owner.PointToScreen(_start);
				WindowInfo window = _windowManager.GetWindowRectFromPoint(screenPoint, new IntPtr[] { _owner.Handle });

				if (window != null)
				{
					//if window is greater screen, we must crop window by screen size
					//window.Bounds.Intersect( Owner.DesktopBounds );
					shotRect = _owner.RectangleToClient(window.Bounds);
					shotRect.Intersect(_owner.RectangleToClient(_owner.Bounds));

				}
			}
			else
			{
				//get region shot
				shotRect = _selectionRect;
			}

			var result = Shot(shotRect);

			Unsubscribe();
			DisposeGdiObjects();

			return result;
		}

		private Image Shot(Rectangle rect)
		{
			if (rect.Width <= 0 || rect.Height <= 0 || rect.X < 0 || rect.Y < 0)
				return null;

			Bitmap bmpCrop = _lightBitmap.Clone(rect, _lightBitmap.PixelFormat);

			if (!string.IsNullOrEmpty(_typedText))
			{
				using (Graphics g = Graphics.FromImage(bmpCrop))
				{
					DrawTypedText(g, new Rectangle(0, 0, rect.Width, rect.Height));
				}
			}

			return bmpCrop;
		}

		#endregion

		private void Unsubscribe()
		{
			_owner.KeyDown -= OnKeyDown;
			_owner.KeyUp -= OnKeyUp;
			_owner.KeyPress -= OnKeyPress;
		}

		private void DisposeGdiObjects()
		{
			_typedTextFontBig.Dispose();
			_typedTextFontSmall.Dispose();
			_storageFont.Dispose();
			_borderPen.Dispose();
			_fillBrush.Dispose();
			_lightBitmap.Dispose();
			var img = _owner.BackgroundImage;
			_owner.BackgroundImage = null;
			img.Dispose();
		}

		private void RecalculateSelection()
		{
			if (_selectionMoveEnabled)
			{
				_selectionRect.X = _startXGreater ? _end.X : _end.X - _selectionRect.Width;
				_selectionRect.Y = _startYGreater ? _end.Y : _end.Y - _selectionRect.Height;

				if (_selectionRect.X < 0)
					_selectionRect.X = 0;

				if (_selectionRect.Y < 0)
					_selectionRect.Y = 0;

				if (_selectionRect.X + _selectionRect.Width > _lightBitmap.Width)
					_selectionRect.X = _lightBitmap.Width - _selectionRect.Width;

				if (_selectionRect.Y + _selectionRect.Height > _lightBitmap.Height)
					_selectionRect.Y = _lightBitmap.Height - _selectionRect.Height;
			}
			else
			{
				_selectionRect.X = Math.Min(_start.X, _end.X);
				_selectionRect.Width = Math.Abs(_start.X - _end.X);

				_selectionRect.Y = Math.Min(_start.Y, _end.Y);
				_selectionRect.Height = Math.Abs(_start.Y - _end.Y);
			}
		}
	}
}
