﻿using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Threading;
using System.Xml.Linq;
using CloudShot.Core;
using CloudShot.Core.Interfaces;
using CloudShot.Core.Types;
using CloudShot.Core.Utils;
using DropNet;
using DropNet.Models;

namespace DropboxImageStorage
{
	public class DropboxStorage : IImageStorage
	{
		private BackgroundWorker _backgroundWorker = null;

		public Image Logo { get { return LocalResources.dropbox; } }
		public string Name { get { return "Dropbox"; } }
		public string Description { get { return "Image will be saved in Dropbox and url will be copied to clipboard"; } }
		public AuthorizationMode AuthorizationMode { get { return AuthorizationMode.OAuth; } }
		public string OAuthTokens { get; set; }
		public bool IsAuthorized
		{
			get { return !string.IsNullOrEmpty(OAuthTokens); }
		}

		public ImageSaveFormat[] SupportedFormats { get { return new[] { ImageSaveFormat.Png, ImageSaveFormat.Jpg, ImageSaveFormat.Jpeg }; } }

		public ImageSaveFormat CurrentImageFormat { get; set; }

		//if you want to add Custom Folder support, just change SaveType to CustomRemote
		public StorageType SupportedStorageType { get { return StorageType.Remote; } }

		//I dont think it's necessary to change destination folder. I keept this code just for example or for the future use.
		private string _folderToSave = @"/";
		public string FolderToSave
		{
			get { return _folderToSave; }
			set { _folderToSave = string.Format("/{0}/", value.Trim('/')); }
		}

		public ShotInfo SaveImageAsync(Image image, string name, Action<ShotInfo> uploadCompleted)
		{
			if (!IsAuthorized)
				throw new UnauthorizedAccessException("You can't take screenshot. You have to authorize first.");

			TokenStorage token = Base64Utils.DeserializeFromBase64(OAuthTokens) as TokenStorage;

			if (token == null)
				throw new UnauthorizedAccessException("Authorization info was corrupted. Please reauthorize in Settings -> SignIn.");

			DropNetClient client = new DropNetClient(DropboxTokens.ApplicationKey, DropboxTokens.ApplicationSecret, token.Key, token.Secret);
			client.UseSandbox = true;

			// build the data stream
			MemoryStream data = new MemoryStream();

			image.Save(data, CurrentImageFormat.GetImageFormat());

			// reset stream
			data.Position = 0;

			// upload data
			MetaData fileInfo = client.UploadFile(FolderToSave, name, data.ToArray());

			string publicUrl = client.GetShare(fileInfo.Path).Url;

			var result = new ShotInfo(name, publicUrl, image);

			if (uploadCompleted != null)
				uploadCompleted(result);

			return result;
		}

		public void SignInAsync(Action<bool> callback)
		{
			DropNetClient client = new DropNetClient(DropboxTokens.ApplicationKey, DropboxTokens.ApplicationSecret);
			string authUrl = client.GetTokenAndBuildUrl();

			Process.Start(authUrl);//open auth url in default web browser

			//TODO маленькая проблема, что если код отмены выполнится позже, чем код авторизации, который написан далее?
			SignOut();//if we was authorized, we have to sign out first

			_backgroundWorker = new BackgroundWorker();
			_backgroundWorker.WorkerSupportsCancellation = true;
			_backgroundWorker.DoWork += OnGetAccessToken;//execute in separate thread

			//execute in main thread
			_backgroundWorker.RunWorkerCompleted += (sender, arg) =>
			{
				if (!arg.Cancelled)
				{
					UserLogin accessToken = arg.Result as UserLogin;
					TokenStorage token = new TokenStorage
					{
						Key = accessToken.Token,
						Secret = accessToken.Secret
					};

					OAuthTokens = Base64Utils.SerializeToBase64(token);

					callback(true);
				}
				else
				{
					callback(false);
				}
			};

			_backgroundWorker.RunWorkerAsync(client);
		}

		public void SignOut()
		{
			// _backgroundWorker is exists only in wait authorization state
			if (_backgroundWorker != null)
			{
				_backgroundWorker.CancelAsync();
				_backgroundWorker = null;
			}

			OAuthTokens = null;
		}

		public XElement[] SaveSettings()
		{
			return new XElement[]
			{
			  new XElement("CurrentImageFormat", CurrentImageFormat),
			  //new XElement("FolderToSave", FolderToSave),
			  new XElement("OAuthTokens", OAuthTokens)
			};
		}

		public void LoadSettings(XElement[] settings)
		{
			if (settings == null || settings.Length == 0)
				return;

			var imageFormat = settings.FirstOrDefault(x => x.Name == "CurrentImageFormat");
			if (imageFormat != null)
				this.CurrentImageFormat = SupportedFormats.FirstOrDefault(x => x.ToString() == imageFormat.Value);

			var tokens = settings.FirstOrDefault(x => x.Name == "OAuthTokens");
			if (tokens != null)
				this.OAuthTokens = tokens.Value;
		}

		private void OnGetAccessToken(object sender, DoWorkEventArgs e)
		{
			const int waitTime = 2000;
			const int repetitions = 150;

			if (e.Cancel)
				return;

			DropNetClient client = e.Argument as DropNetClient;

			if (client != null)
			{
				UserLogin accessToken = null;

				//https://www.dropbox.com/developers/start/authentication#python
				//We don't want anyone other than the user who requested access using this URL, 
				//which is why the request tokens are only live for five minutes. 
				//If the user sits on this link for greater than five minutes, 
				//we'll politely request they return to your app to try again.
				for (int i = 0; i < repetitions; ++i)
				{
					if (((BackgroundWorker)sender).CancellationPending)
					{
						e.Cancel = true;
						break;
					}

					Thread.Sleep(waitTime);
					try
					{
						accessToken = client.GetAccessToken();
						if (accessToken != null)
							break;
					}
					catch { }
				}

				e.Result = accessToken;
			}
		}
	}

	[Serializable]
	public class TokenStorage
	{
		public string Key { get; set; }
		public string Secret { get; set; }
	}
}